// MPanel.java

/*
  This software is part of the Gidlet framework.
  It is Open Source Free Software, so you may
    - run the code for any purpose
    - study how the code works and adapt it to your needs
    - integrate all or parts of the code in your own programs
    - redistribute copies of the code
    - improve the code and release your improvements to the public
  However the use of the code is entirely your responsibility.
 */


/*
 * Window coordinates:  (wx, wy)
 * Pixel coordinates: (px, py) (display pixels)
 * Tranformation: px = px( wx ), py = py( wy )
 * Pixel range: 0..xPix, 0..yPix (0,0) upper left corner, x to right, y down
 * Window range: xmin..xmax, ymin..ymax (0,0) lower left corner, x to right, y up
 * Display width in pixels (#nb of pixels): width, display height in pixels: height
 * xPix = width - 1, yPix = height - 1
 * e.g. width = 240, height = 291 (displayable display resolution 240x291)
 * Pixel xrange: 0..239, yrange: 0..290
 * px = a*wx + b
 * py = c*wy + d
 * with a = xPix / ( xmax - xmin )
 * b = xPix * xmin / ( xmin - xmax )
 * c = yPix / ( ymin - ymax )
 * d = yPix * ymax / ( ymax - ymin )
 * Inverse:
 * wx = (px - b) /a
 * wy = (py - d) /c
 */

package ch.aplu.gidlet;

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.*;
import java.io.*;
import java.util.*;
import ch.aplu.gidlet.*;

/**
 * Simple graphics window of full display width and height (client drawing area)
 * using a coordinate system with x-axis from left to right, y-axis from bottom to top
 * (called window coordinates, default range 0..1, 0..1).
 * The drawing methods perform all drawings in an offscreen buffer
 * and automatically flush it on the display, so the graphics is shown step-by-step.
 * <br><br>
 *
 * A current graph position is used, that remembers the end position of the last
 * drawing process and where new drawings start.<br><br>
 *
 * If pixel accuracy is needed, use the graphcis methods starting with _ <br><br>
 *
 * A soft button 'Exit' is shown. Hitting the corresponding key will
 * call the Gidlet's doExit(). If you do not need this button,
 * invoke removeExitButton().<br><br>
 * If you implement your own CommandListener and call setCommandListener(), the internal
 * CommandListener is disabled.<br><br>
 *
 * To avoid flickering in animated graphics, turn automatic flushing off by calling
 * enableFlush(false). The drawing methods only modifies the offscreen buffer now. Call
 * flushGraphics() to render it to the display.<br><br>
 *
 * main() serves as entry point for the user programm. It is called in the run-method
 * of a separate thread and should terminate to avoid non-terminated threads.<br><br>
 *
 * An instance of MConsole may be created when declaring instance variable of the application
 * program. To garantee that the initialization is executed in the Gidlet's thread,
 * these instances are queued and the intialisation takes place in the same order
 * they are created when main() is started.<br><br>
 *
 * Example:<p>
 * <code><pre>
 * import ch.aplu.gidlet.*;
 * public class GidletEx3 extends Gidlet
 * {
 *   public void main()
 *   {
 *     MPanel mp = new MPanel(this, "Chessboard");
 *     mp.window(0, 8, 0, 8);
 *     mp.rectangle(0, 0, 8, 8);
 *      for (int i = 0; i < 8; i++)
 *        for (int j = 0; j < 8; j++)
 *          if ((i + j) % 2 == 0)
 *            mp.fillRectangle(i, j, i + 1, j + 1);
 *    }
 *  }
 * </pre></code>
 *<br>
 * More than one instance may be created. The constructor flag visible
 * may be used, to select if the display is initially shown or hidden.
 * A hidden display will become visible by calling show(). There is no other way to
 * make it hidden than showing a different display. All graphics methods may
 * be called while the display is hidden.
 */

public class MPanel extends GameCanvas
{
// ------------------ Inner class MyCommandListener -------------
private class MyCommandListener implements CommandListener
{
  public void commandAction(Command c, Displayable d)
  {
    if (c == exitCommand)
      gidlet.doExit();
    if (c == okCommand)
    {
      gidlet.doOk();
      gidlet.doOk(d);
      if (gidlet.isSleeping())
        gidlet.wakeUp();
    }
  }
}

// ------------------ Inner class TimerTask ---------------------
private class MyTimerTask extends TimerTask
{
  public void run()
  {
    waitVisible = false;
  }
}

// ------------------ End of inner classes ----------------------

  private Gidlet gidlet;
  private Command exitCommand = new Command("Exit", Command.EXIT, 1);
  private Command okCommand = null;

  private Image saveImage;
  private Graphics saveGraphics;
  private String title;
  private int xPix;
  private int yPix;
  private double xmin;
  private double xmax;
  private double ymin;
  private double ymax;
  private boolean visible;
  private int xCurrent;
  private int yCurrent;
  private double a, b, c, d;
  private int width;
  private int height;
  private int bgColor;
  private int penColor;
  private boolean isFlush = true;
  private Graphics g;
  private KeyListener keyListener = null;
  private Hashtable textLabels = new Hashtable();
  private boolean waitVisible;

  // ============== Constructors =========================================

  /**
   * Constructs a MPanel and show the display with given title
   * and given window coordinates,
   * (0, 0) is at lower left corner which is the current graph cursor position.
   * The parameter visible controls, if the display is initially shown.
   */
  public MPanel(String title, double xmin, double xmax, double ymin, double ymax, boolean visible)
  {
    super(false);
    this.gidlet = Gidlet.myGidlet;
    this.title = title;
    this.xmin = xmin;
    this.xmax = xmax;
    this.ymin = ymin;
    this.ymax = ymax;
    this.visible = visible;
    if (Thread.currentThread().getName().equals("Gidlet-Thread"))
      init();
    else // defer initialization
      (gidlet.displays).addElement(this);
  }

  /**
   * Same as MPanel(title, xmin, xmax, ymin, ymax, visible) with visible = true.
   */
  public MPanel(String title, double xmin, double xmax, double ymin, double ymax)
  {
    this(title, xmin, xmax, ymin, ymax, true);
  }

  /**
   * Constructs a MPanel and show the display with no title.
   * given window coordinates.
   * (0, 0) is at lower left corner which is the current graph cursor position.
   * The parameter visible controls, if the display is initially shown.
   */
  public MPanel(double xmin, double xmax, double ymin, double ymax, boolean visible)
  {
    this("", xmin, xmax, ymin, ymax, visible);
  }

  /**
   * Same as MPanel(xmin, xmax, ymin, ymax, visible) with visible = true.
   */
  public MPanel(double xmin, double xmax, double ymin, double ymax)
  {
    this(xmin, xmax, ymin, ymax, true);
  }


  /**
   * Constructs a MPanel and show the display with given title.
   * Window coordinates are 0, 1, 0, 1.
   * (0, 0) is at lower left corner which is the current graph cursor position.
   * The parameter visible controls, if the display is initially shown.
   */
  public MPanel(String title, boolean visible)
  {
    this(title, 0, 1, 0, 1, visible);
  }

  /**
   * Same as MPanel(title, visible) with visible = true.
   */
  public MPanel(String title)
  {
    this(title, true);
  }


  /**
   * Constructs a MPanel and show the window with no title.
   * Window coordinates are 0, 1, 0, 1.
   * (0, 0) is at lower left corner which is the current graph cursor position.
   * The parameter visible controls, if the display is initially shown.
   */
  public MPanel(boolean visible)
  {
    this("", 0, 1, 0, 1, visible);
  }

  /**
   * Same as MPanel(visible) with visible = true.
   */
  public MPanel()
  {
    this(true);
  }
  // ============== End of constructors ===================================

  /**
   * Initializes MPanel. Called by main() if MPanel is created before main() is called.
   */
  protected void init()
  {
    addCommand(exitCommand);
    setCommandListener(new MyCommandListener());
    g = getGraphics();
    height = getHeight();
    width = getWidth();
    window(xmin, xmax, ymin, ymax);
    bgColor(0XFFFFFF); // white
    color(0); // black
    xCurrent = toPixelX(0);
    yCurrent = toPixelY(0);
    if (visible)
      show(true);
  }


  /**
   * Shows the form display if it is not yet visible or
   * it was previously hidden by another display.
   * Because it is not garanteed that the diplay is actually visible when
   * Display.setCurrent() returns, wait for a maximum of 5 s for the display
   * to get shown.
   *
   */
   public void show()
   {
     show(true);
   }

  /**
   * Shows the graphics display if it was previously hidden by another display
   * or not yet displayed.<br>
   * If wait = true, block until the display becomes actually visible, up
   * to a maximum auf 5 seconds. This may be useful to observe the start of
   * an animation.
   */
  public void show(boolean wait)
  {
    gidlet.display.setCurrent(this);
    if (wait)
    {
      waitVisible = true;
      Timer t = new Timer();
      t.schedule(new MyTimerTask(), 5000);
      while (waitVisible && !isShown())
        Gidlet.delay(100);
      t.cancel();
    }
    flushGraphics();
  }

  /**
   * Sets the background color. All drawing are erased.
   * Returns the previous color.
   */
  public int bgColor(int color)
  {
    int oldBgColor = bgColor;
    int oldPenColor = penColor;
    bgColor = color;
    g.setColor(bgColor);
    g.fillRect(0, 0, width, height);
    g.setColor(oldPenColor);
    g.drawString(title, getWidth()/2, 5,
                   Graphics.TOP | Graphics.HCENTER);
    if (isFlush)
      flushGraphics();
    return oldBgColor;
  }

  /**
   *  Returns the current background color.
   */
  public int getBgColor()
  {
    return bgColor;
  }

 /**
   *  Returns the current drawing color.
   */
  public int getColor()
  {
    return penColor;
  }

  /**
   * Enables or disables the automatic flush in graphics methods.
   * Returns the current state of flushing.<br>
   * When automatic flush is disabled, the method flushGraphics() must be called to show
   * the image on the display.<br>
   * Useful to avoid flickering while animating an image. If flush is disabled
   * clear() will only erase the offscreen buffer and not the display. flushGraphics() should
   * be called after drawing the new situation in order to copy the offscreen buffer to the display.<br>
   */
  public boolean enableFlush(boolean doFlush)
  {
    boolean oldFlush = isFlush;;
    isFlush = doFlush;
    return oldFlush;
  }

  /**
   * Clears the graphics window (fully paint with background color)
   * (and the offscreen buffer used by the window).
   * Sets the current graph position to window coordiantes (0, 0).<br>
   * If enableFlush(false) only clear the offscreen buffer.
   */
  public void clear()
  {
    g.setColor(bgColor);
    g.fillRect(0, 0, width, height);
    g.setColor(penColor);
    xCurrent = toPixelX(0);
    yCurrent = toPixelY(0);
    if (isFlush)
      flushGraphics();
  }

  /**
   * Same as clear() but let the current graph position unchanged.
   */
  public void erase()
  {
    g.setColor(bgColor);
    g.fillRect(0, 0, width, height);
    g.setColor(penColor);
    if (isFlush)
      flushGraphics();
  }

  /**
   * Draws a circle with center at the current graph position
   * and given radius in pixel coordinates.
   * The graph position is unchanged.
   */
  public void _circle(int radius)
  {
    int ulx = xCurrent - radius;
    int uly = yCurrent - radius;
    g.drawArc(ulx, uly, 2*radius, 2*radius, 0, 360);
    if (isFlush)
      flushGraphics();
  }

  /**
   * Draws a circle with center at the current graph position
   * and given radius in horizontal window coordinates.
   * The graph position is unchanged.
   */
  public void circle(double radius)
  {
   // adapt to horizontal to draw circle and not ellipse
    int rx = toPixelWidth(radius);
    _circle(rx);
  }

  /**
   * Draws a filled circle with center at the current graph position
   * and given radius in pixel coordinates.
   * The graph position is unchanged.
   */
  public void _fillCircle(int radius)
  {
    int ulx = xCurrent - radius;
    int uly = yCurrent - radius;
    g.fillArc(ulx, uly, 2*radius, 2*radius, 0, 360);
    if (isFlush)
      flushGraphics();
  }

  /**
   * Draws a filled circle with center at the current graph position
   * and given radius in horizonal window coordinates.
   * The graph position is unchanged.
   */
  public void fillCircle(double radius)
  {
   // adapt to horizontal to draw circle and not ellipse
    int rx = toPixelWidth(radius);
    _fillCircle(rx);
  }

  /**
   * Draws an elliptical arc inside bounding rectangle centered at current graph position
   * with given width and height in pixel coordinates.
   * The arc starts at given startAngle and covers the given arcAngle (both in degrees).
   * The graph position is unchanged.
   */
  public void _arc(int width, int height, int startAngle, int arcAngle)
  {
    int ulx = xCurrent - width/2;
    int uly = yCurrent - height/2;
    g.drawArc(ulx, uly, width, height, startAngle, arcAngle);
    if (isFlush)
      flushGraphics();
  }

  /**
   * Draws an elliptical arc inside bounding rectangle centered at current graph position
   * with given width and height in window coordinates.
   * The arc starts at given startAngle (with respect to east direction)
   * and covers the given arcAngle (counter-clockwise) (both angles in degrees).
   * The graph position is unchanged.
   */
  public void arc(double width, double height, int startAngle, int arcAngle)
  {
    _arc(toPixelWidth(width), toPixelHeight(height), startAngle, arcAngle);
  }

  /**
   * Draws a filled elliptical arc inside bounding rectangle centered at current graph position
   * with given width and height in pixel coordinates.
   * The arc starts at given startAngle and covers the given arcAngle (both in degrees).
   * The graph position is unchanged.
   */
  public void _fillArc(int width, int height, int startAngle, int arcAngle)
  {
    int ulx = xCurrent - width/2;
    int uly = yCurrent - height/2;
    g.fillArc(ulx, uly, width, height, startAngle, arcAngle);
    if (isFlush)
      flushGraphics();
  }

  /**
   * Draws a filled elliptical arc inside bounding rectangle centered at current graph position
   * with given width and height in window coordinates.
   * The arc starts at given startAngle (with respect to east direction)
   * and covers the given arcAngle (counter-clockwise) (both angles in degrees).
   * The graph position is unchanged.
   */
  public void fillArc(double width, double height, int startAngle, int arcAngle)
  {
    _fillArc(toPixelWidth(width), toPixelHeight(height), startAngle, arcAngle);
  }

  /**
   * Sets the current color of the drawing pen.
   * Returns the previous color.
   */
  public int color(int color)
  {
    int oldPenColor = penColor;
    penColor = color;
    g.setColor(color);
    return oldPenColor;
  }

  /**
   * Sets the current color of the drawing pen in RGB-format.
   * Returns the previous color.
   */
  public int color(int r, int g, int b)
  {
    int rgb = (r << 16) + (g << 8) + b;
    return color(rgb);
  }

  /**
   * Shortcut for flushGraphics()
   */
  public void flush()
  {
    flushGraphics();
  }

  /**
   * Draws the given title with current penColor centered in top area of display.
   * Does not erase previous text at same position.
   */
  public void title(String text)
  {
    g.drawString(text, getWidth()/2, 5,
                   Graphics.TOP | Graphics.HCENTER);
    if (isFlush)
      flushGraphics();
  }

  /**
   * Draws the given text with current drawing color and font left adjusted
   * at given y pixel coordinate.
   * Erases previous text at same position.
   * To position the text horizontally, use leading spaces.
   * @see #color
   * @see #setFont
   */
  public void _label(String text, int y)
  {
    String oldText = (String)textLabels.get(new Integer(y));
    if (oldText != null)
    {
      int oldColor = color(bgColor);
      g.drawString(oldText, 1, y,
                   Graphics.TOP | Graphics.LEFT);
      color(oldColor);
    }

    g.drawString(text, 1, y,
                   Graphics.TOP | Graphics.LEFT);
    textLabels.put(new Integer(y), text);
    if (isFlush)
      flushGraphics();
  }

  /**
   * Draws the given text with current drawing color and font left adjusted
   * at given window y coordinate.
   * Does not erase previous text at same position.
   * To position the text horizontally, use leading spaces.
   * @see #color
   * @see #setFont
   */
  public void label(String text, double y)
  {
    _label(text, toPixelY(y));
  }

  /**
   * Draws a single point at the given pixel coordinates.
   * The graph position is set to the given point.
   */
  public void _point(int x, int y)
  {
    g.drawLine(x, y, x, y);
    xCurrent = x;
    yCurrent = y;
    if (isFlush)
      flushGraphics(x - 2, y - 2, 4, 4);
  }

  /**
   * Draws a single point at the given window coordinates.
   * The graph position is set to the given point.
   */
  public void point(double x, double y)
  {
    _point(toPixelX(x), toPixelY(y));
  }

  /**
   * Draws a line with given pixel coordinates
   * and sets the graph position to the endpoint.
   */
  public void _line(int x1, int y1, int x2, int y2)
  {
    g.drawLine(x1, y1, x2, y2);
    xCurrent = x2;
    yCurrent = y2;
    if (isFlush)
      flushGraphics();
  }

  /**
   * Draws a line with given window coordinates
   * and sets the graph position to the endpoint.
   */
  public void line(double x1, double y1, double x2, double y2)
  {
    int xStart = toPixelX(x1);
    int yStart = toPixelY(y1);
    int xEnd = toPixelX(x2);
    int yEnd = toPixelY(y2);
    _line(xStart, yStart, xEnd, yEnd);
  }

  /**
   * Draws a line from current graph position to given pixel coordinates and
   * sets the graph position to the endpoint.
   */
  public void _draw(int x, int y)
  {
    _line(xCurrent, yCurrent, x, y);
  }

  /**
   * Draws a line from current graph position to given window coordinates and
   * sets the graph position to the endpoint.
   */
  public void draw(double x, double y)
  {
    _draw(toPixelX(x), toPixelY(y));
  }

  /**
   * Sets the current graph position to given pixel coordinates
   * (without drawing anything).
   */
  public void _move(int x, int y)
  {
    xCurrent = x;
    yCurrent = y;
  }

  /**
   * Sets the current graph position to given window coordinates
   * (without drawing anything).
   */
  public void move(double x, double y)
  {
    _move(toPixelX(x), toPixelY(y));
  }

  /**
   * Draws a triangle with given corners in pixel coordintates.
   * The graph position is unchanged.
   */
  public void _triangle(int x1, int y1,
                        int x2, int y2,
                        int x3, int y3)
  {
    int oldX = xCurrent;
    int oldY = yCurrent;
    _move(x1, y1);
    _draw(x2, y2);
    _draw(x3, y3);
    _draw(x1, y1);
    xCurrent = oldX;
    yCurrent = oldY;
    if (isFlush)
      flushGraphics();
  }

  /**
   * Draws a triangle with given corners in window coordintates.
   * The graph position is unchanged.
   */
  public void triangle(double x1, double y1,
                       double x2, double y2,
                       double x3, double y3)
  {
    _triangle(toPixelX(x1), toPixelY(y1),
              toPixelX(x2), toPixelY(y2),
              toPixelX(x3), toPixelY(y3));
  }

  /**
   * Draws a filled triangle with given corners in pixel coordintates
   * using the current color.
   * The graph position is unchanged.
   */
  public void _fillTriangle(int x1, int y1,
                            int x2, int y2,
                            int x3, int y3)
  {
    g.fillTriangle(x1, y1, x2, y2, x3, y3);
    if (isFlush)
      flushGraphics();
  }

  /**
   * Draws a filled triangle with given corners in window coordintates
   * using the current color.
   * The graph position is unchanged.
   */
  public void fillTriangle(double x1, double y1,
                           double x2, double y2,
                           double x3, double y3)
  {
    _fillTriangle(toPixelX(x1), toPixelY(y1),
                  toPixelX(x2), toPixelY(y2),
                  toPixelX(x3), toPixelY(y3));
  }

  /**
   * Draws a rectangle with center at the current graph position
   * and given width and height in pixel coordinates.
   * The graph position is unchanged.
   */
  public void _rectangle(int width, int height)
  {
    int anchorx = xCurrent - Math.abs(width) / 2;
    int anchory = yCurrent - Math.abs(height) / 2;
    g.drawRect(anchorx, anchory, Math.abs(width), Math.abs(height));
    if (isFlush)
      flushGraphics();
  }

 /**
   * Draws a rectangle with center at the current graph position
   * and given width and height in window coordinates.
   * The graph position is unchanged.
   */
  public void rectangle(double width, double height)
  {
    _rectangle(toPixelWidth(width), toPixelHeight(height));
  }

  /**
   * Draws a rectangle with given opposite corners
   * in pixel coordinates.
   * The graph position is unchanged.
   */
  public void _rectangle(int x1, int y1, int x2, int y2)
  {
    int width = Math.abs(x2 - x1);
    int height = Math.abs(y2 - y1);
    int anchorx = Math.min(x1, x2);
    int anchory = Math.min(y1, y2);
    g.drawRect(anchorx, anchory, width, height);
    if (isFlush)
      flushGraphics();
  }

   /**
   * Draws a rectangle with given opposite corners
   * in window coordinates.
   * The graph position is unchanged.
   */
   public void rectangle(double x1, double y1, double x2, double y2)
  {
    _rectangle(toPixelX(x1), toPixelY(y2), toPixelX(x2), toPixelY(y1));
  }


 /**
   * Draws a filled rectangle with center at the current graph position
   * and given width and height in window coordinates using the current color.
   * The graph position is unchanged.
   */
  public void _fillRectangle(int width, int height)
  {
    int anchorx = xCurrent - Math.abs(width) / 2;
    int anchory = yCurrent - Math.abs(height) / 2;
    g.fillRect(anchorx, anchory, Math.abs(width), Math.abs(height));
    if (isFlush)
      flushGraphics();
  }

 /**
   * Draws a filled rectangle with center at the current graph position
   * and given width and height in window coordinates using the current color.
   * The graph position is unchanged.
   */
  public void fillRectangle(double width, double height)
  {
    _fillRectangle(toPixelWidth(width), toPixelHeight(height));
  }

  /**
   * Draws a filled rectangle with given opposite corners
   * in pixel coordinates.
   * The graph position is unchanged.
   */
  public void _fillRectangle(int x1, int y1, int x2, int y2)
  {
    int width = Math.abs(x2 - x1);
    int height = Math.abs(y2 - y1);
    int anchorx = Math.min(x1, x2);
    int anchory = Math.min(y1, y2);
    g.fillRect(anchorx, anchory, width, height);
    if (isFlush)
      flushGraphics();
  }

   /**
   * Draws a filled rectangle with given opposite corners
   * in window coordinates.
   * The graph position is unchanged.
   */
   public void fillRectangle(double x1, double y1, double x2, double y2)
  {
    _fillRectangle(toPixelX(x1), toPixelY(y2), toPixelX(x2), toPixelY(y1));
  }

  /**
   * Sets window coordinates.
   * left_x, right_x, bottom_y, top_y
   */
  public void window(double xmin, double xmax, double ymin, double ymax)
  {
    double ytop = ymin + getAspectRatio()*(ymax - ymin);
    this.xmin = xmin;
    this.xmax = xmax;
    this.ymin = ymin;
    this.ymax = ytop;
    xPix = width - 1;
    yPix = height - 1;

    a = xPix / (this.xmax - this.xmin);
    b = xPix * this.xmin / (this.xmin - this.xmax);
    c = yPix / (this.ymin - this.ymax);
    d = yPix * this.ymax / (this.ymax - this.ymin);

    xCurrent = toPixelX(0);
    yCurrent = toPixelY(0);
   }

  /**
   * Converts window coordinates to pixel coordinates (horizontal).
   */
  public int toPixelX(double windowX)
  {
    return rint(a * windowX + b);     // Rounding
  }

  /**
   * Converts window coordinates to pixel coordinates (vertical).
   */
  public int toPixelY(double windowY)
  {
    return rint(c * windowY + d);     // Rounding
  }

  /**
   * Converts window coordinates increment to pixel coordinates increment (horizontal).
   * Increment is always positive.
   */
  public int toPixelWidth(double windowWidth)
  {
    return rint(Math.abs(a * windowWidth));
  }

  /**
   * Converts window coordinates increment to pixel coordinates increment (vertical).
   * Increment is always positive.
   */
  public int toPixelHeight(double windowHeight)
  {
    return rint(Math.abs(c * windowHeight));
  }

  /**
   * Converts pixel coordinates to window coordinates (horizontal).
   */
  public double toWindowX(int userX)
  {
    return (userX - b) / a;
  }

  /**
   * Converts pixel coordinates to window coordinates (vertical).
   */
  public double toWindowY(int userY)
  {
    return (userY - d) / c;
  }

  /**
   * Converts pixel coordinates increment to window coordinates increment (horizontal).
   * Increment is always positive.
   */
  public double toWindowWidth(int pixelWidth)
  {
    return Math.abs(pixelWidth / a);
  }

  /**
   * Converts pixel coordinates increment to window coordinates increment (vertical).
   * Increment is always positive.
   */
  public double toWindowHeight(int pixelHeight)
  {
    return Math.abs(pixelHeight / c);
  }

 /**
  * Returns the ratio of canvas height to width.
  */
  public double getAspectRatio()
  {
    return (height / (double)width);
  }

 /**
  * Returns the graphics context.
  */
  public Graphics getG()
  {
    return g;
  }

  /**
   * Returns pixel coordinate x of current graph position.
   */
  public int _getPosX()
  {
    return xCurrent;
  }

  /**
   * Returns window coordinate x of current graph position.
   */
  public double getPosX()
  {
    return toWindowX(xCurrent);
  }

  /**
   * Returns pixel coordinate y of current graph position.
   */
  public int _getPosY()
  {
    return yCurrent;
  }

  /**
   * Returns window coordinate y of current graph position.
   */
  public double getPosY()
  {
    return toWindowY(yCurrent);
  }

  /**
   * Registers an KeyListener to get a notification when a button is clicked.
   */
  public void addKeyListener(KeyListener keyListener)
  {
    this.keyListener = keyListener;
  }

  /**
   * Removes a registered KeyListener.
   */
  public void removeKeyListener()
  {
    keyListener = null;
  }

  /**
   * Loads image from given resource and returns it's Image reference.
   * Returns null, if loading fails.
   */
  public Image getImage(String imageUrl)
  {
    InputStream is = getClass().getResourceAsStream(imageUrl);
    Image img;
    try
    {
      img = Image.createImage(is);  // Blocking until read and decoded
      if (is != null)
        is.close();
    }
    catch (IOException ex)
    {
      return null;
    }
    return img;
  }

  /**
   * Loads image from given resource and returns it's Sprite reference.
   * Returns null, if loading fails.<br>
   */
  public Sprite getSprite(String imageUrl)
  {
    Image spriteImg = getImage(imageUrl);
    if (spriteImg == null)
      return null;
    else
      return new Sprite(spriteImg);
  }

  /**
   * Loads image from given resource with given frameWidth, frameHeight
   * and returns it's Sprite reference.
   * Returns null, if loading fails.<br>
   */
  public Sprite getSprite(String imageUrl, int frameWidth, int frameHeight)
  {
    Image spriteImg = getImage(imageUrl);
    if (spriteImg == null)
      return null;
    else
      return new Sprite(spriteImg, frameWidth, frameHeight);
  }

   /**
   * Draws the given sprite at given pixel coordinates
   * (upper left corner of image).
   * If sprite == null, nothing happens.
   * Image transparancy is supported. Transparent parts must be
   * created by editing the png-file.
   */
  public void _drawSprite(Sprite sprite, int x, int y)
  {
    if (sprite != null)
    {
      sprite.setPosition(x, y);
      sprite.paint(g);
      if (isFlush)
        flushGraphics();
    }
  }

 /**
   * Draws the given sprite at given window coordinates
   * (upper left corner of image).
   * If sprite == null, nothing happens.
   * Image transparancy is supported. Transparent parts must be
   * created by editing the png-file.
   */
  public void drawSprite(Sprite sprite, double x, double y)
  {
    _drawSprite(sprite, toPixelX(x), toPixelY(y));
  }

  /**
   * Shows the PNG image from given file url at given pixel coordinates
   * (upper left corner of image).
   * Returns true if successful.
   * If enableFlush(false), the image is drawn in offscreen buffer only.
   * Image transparancy is supported. Transparent parts must be
   * created by editing the png-file.
   */
  public boolean _drawImage(String imageUrl, int x, int y)
  {
    InputStream is = getClass().getResourceAsStream(imageUrl);
    Image img;
    try
    {
      img = Image.createImage(is);  // Blocking until read and decoded
      if (is != null)
        is.close();
    }
    catch (IOException ex)
    {
      return false;
    }
    int imgHeight = img.getHeight();
    int imgWidth = img.getWidth();
    g.drawImage(img, x, y, Graphics.TOP | Graphics.LEFT);
    if (isFlush)
      flushGraphics(x, y, imgWidth, imgHeight);
    return true;
  }

  /**
   * Shows the PNG image from given file url at given window coordinates
   * (upper left corner of image).
   * Returns true if successful.
   * If enableFlush(false), the image is drawn in offscreen buffer only.
   * Image transparancy is supported. Transparent parts must be
   * created by editing the png-file.
   */
  public boolean drawImage(String imageUrl, double x, double y)
  {
    return _drawImage(imageUrl, toPixelX(x), toPixelY(y));
  }

  /**
   * Returns the width (horizontal size) of the PNG image from the given url (in pixel coordinates).
   * Returns 0, if PNG image is invalid.
   */
  public int _imageWidth(String imageUrl)
  {
    InputStream is = getClass().getResourceAsStream(imageUrl);
    Image img;
    try
    {
      img = Image.createImage(is);
      if (is != null)
        is.close();
    }
    catch (IOException ex)
    {
      return 0;
    }
    return img.getWidth();
  }

  /**
   * Returns the width (horizontal size) of the PNG image from the given url (in window coordinates).
   * Returns 0, if PNG image is invalid.
   */
  public double imageWidth(String imageUrl)
  {
    int rc = _imageWidth(imageUrl);
    if (rc == 0)
      return 0;
    else
      return toWindowWidth(rc);
  }

  /**
   * Returns the height (vertical size) of the PNG image from the given url (in pixel coordinates).
   * Returns 0, if PNG image is invalid.
   */
  public int _imageHeight(String imageUrl)
  {
    InputStream is = getClass().getResourceAsStream(imageUrl);
    Image img;
    try
    {
      img = Image.createImage(is);
      if (is != null)
       is.close();
    }
    catch (IOException ex)
    {
      return 0;
    }
    return img.getHeight();
  }

  /**
   * Returns the height (vertical size) of the PNG image from the given url (in window coordinates).
   * Returns 0, if PNG image is invalid.
   */
  public double imageHeight(String imageUrl)
  {
    int rc = _imageHeight(imageUrl);
    if (rc == 0)
      return 0;
    else
      return toWindowHeight(rc);
  }

  /**
   * Returns width of display in pixel coordinates.
   * Pixel coordinates 0 up to this value (inclusive)
   * Number of pixels: this value + 1
   */
  public int _displayWidth()
  {
    return xPix;
  }

  /**
   * Return numbers of pixels in horizontal direction.
   */
  public int displayResolutionX()
  {
    return width;
  }

  /**
   * Returns width of display in window coordinates.
   */
  public double displayWidth()
  {
    return toWindowWidth(xPix);
  }

  /**
   * Returns height of display in pixel coordinates.
   * Pixel coordinates 0 up to this value (inclusive)
   * Number of pixels: this value + 1
   */
  public int _displayHeight()
  {
    return yPix;
  }

  /**
   * Returns number of pixels in vertical direction.
   */
  public int displayResolutionY()
  {
    return height;
  }

  /**
   * Returns height of display in window coordinates.
   */
  public double displayHeight()
  {
    return toWindowHeight(yPix);
  }

  /**
   * Sets the current text font for drawing labels.
   * If the font is changed, existing text is not affected.
   * Default font is FACE_SYSTEM, STYLE_PLAIN, SIZE_MEDIUM.
   * @param face one of Font.FACE_SYSTEM, Font.FACE_MONOSPACE, or Font.FACE_PROPORTIONAL<br>
   * @param style Font.STYLE_PLAIN, or a combination of Font.STYLE_BOLD, Font.STYLE_ITALIC, and Font.STYLE_UNDERLINED<br>
   * @param size one of Font.SIZE_SMALL, Font.SIZE_MEDIUM, or Font.SIZE_LARGE
   */
  public void setFont(int face, int style, int size)
  {
    Font font = Font.getFont(face, style, size);
    g.setFont(font);
  }

  // Rounding doubles
  private int rint(double x)
  {
    if (x > 0)
      return (int)(x + 0.5);
    else
      return (int)(x - 0.5);
   }

  /**
   * Used to get key events.
   */
  protected void keyPressed(int keyCode)
  {
    if (keyListener != null)
      keyListener.keyPressed(keyCode);
  }

  /**
   * Used to get key events.
   */
  protected void keyReleased(int keyCode)
  {
    if (keyListener != null)
      keyListener.keyReleased(keyCode);
  }

  /**
   * Used to get key events.
   */
  protected void keyRepeated(int keyCode)
  {
    if (keyListener != null)
      keyListener.keyRepeated(keyCode);
  }

  /**
   * Saves the current graphics content to an internal offscreen buffer.
   * Useful to move another graphics object on this background
   * image by repetive call of restorePanel() and graphics commands.
   * @see #restorePanel
   */
  public void savePanel()
  {
    saveImage = Image.createImage(width, height);
    saveGraphics = saveImage.getGraphics();
    paint(saveGraphics);
  }

  /**
   * Renders the content of the internal offscreen buffer.
   * If no image has been stored nothing happens.
   * @see  #savePanel
   */
  public void restorePanel()
  {
    if (saveImage != null)
    {
      g.drawImage(saveImage, 0, 0, 0);
      flushGraphics();
    }
  }

  /**
   * Adds an OK soft button with label "OK". When pressed, the
   * Gidlet's callback notification doOk() is called unless you
   * register your own CommandListener.
   */
  public void addOkButton()
  {
    addOkButton("OK");
  }

  /**
   * Adds an OK soft button with given label. When pressed, the
   * Gidlet's callback notification doOk() is called unless you
   * register your own CommandListener.
   */
  public void addOkButton(String label)
  {
    okCommand = new Command(label, Command.OK, 1);
    addCommand(okCommand);
  }

  /**
   * Removes Exit soft button.
   * Gidlet's callback notification doExit() is disabled.
   */
  public void removeExitButton()
  {
   removeCommand(exitCommand);
  }

  /**
   * Removes Ok soft button.
   * Gidlet's callback notification doOk() is disabled.
   */
  public void removeOkButton()
  {
   removeCommand(okCommand);
  }
}


