// MTextBox.java

/*
  This software is part of the Gidlet framework.
  It is Open Source Free Software, so you may
    - run the code for any purpose
    - study how the code works and adapt it to your needs
    - integrate all or parts of the code in your own programs
    - redistribute copies of the code
    - improve the code and release your improvements to the public
  However the use of the code is entirely your responsibility.
*/

package ch.aplu.gidlet;

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import java.util.Hashtable;
import java.util.*;
import ch.aplu.gidlet.*;

/**
 * TextBox display for full screen input and output. Text may be edited using the
 * platform's editing capabilities.<br><br>
 *
 * A soft button 'Exit' is shown. Hitting the corresponding key will
 * call the Gidlet's doExit(). If you do not need this button,
 * invoke removeExitButton().<br><br>
 * If you implement your own CommandListener and call setCommandListener(), the internal
 * CommandListener is disabled.<br><br>
 *
 * main() serves as entry point for the user programm. It is called in the run-method
 * of a separate thread and should terminate to avoid non-terminated threads.<br><br>
 *
 * Creation may take place within main() or when declaring instance variables
 * of the application program.
 */
public class MTextBox extends TextBox
{
// ------------------ Inner class MyCommandListener -------------
private class MyCommandListener implements CommandListener
{
  public void commandAction(Command c, Displayable d)
  {
    if (c == exitCommand)
      gidlet.doExit();
    if (okCommand != null && c == okCommand)
    {
      gidlet.doOk();
      gidlet.doOk(d);
      if (gidlet.isSleeping())
        gidlet.wakeUp();
    }
  }
}

// ------------------ Inner class TimerTask ---------------------
private class MyTimerTask extends TimerTask
{
  public void run()
  {
    waitVisible = false;
  }
}
// ------------------ End of inner classes ----------------------

  private Gidlet gidlet;
  private Command exitCommand = new Command("Exit", Command.EXIT, 1);
  private Command okCommand = null;
  private boolean waitVisible;
  private int maxSize;

  /**
   * Constructs a MTextBox with given title and given initial text.
   * The display is initially invisible and should be made visible
   * by calling show().
   */
  public MTextBox(String title, String text)
  {
     // Initial size: 1000
    super(title, null, 1000, TextField.ANY);
    maxSize = getMaxSize();
    setText(text);
    this.gidlet = Gidlet.myGidlet;
    addCommand(exitCommand);
    setCommandListener(new MyCommandListener());
  }

  /**
   * Constructs a MTextBox with given title and no initial text.
   * The display is initially invisible and should be made visible
   * by calling show().
   */
  public MTextBox(String title)
  {
    this(title, "");
  }

  /**
   * Shows the form display if it is not yet visible or
   * it was previously hidden by another display.
   * Because it is not garanteed that the diplay is actually visible when
   * Display.setCurrent() returns, wait for a maximum of 5 s for the display
   * to get shown.
   *
   */
  public void show()
  {
    show(true);
  }

  /**
   * Shows the form display if it is not yet visible or
   * it was previously hidden by another display.
   * If wait = true, block until the display becomes actually visible, up
   * to a maximum auf 5 seconds. This may be useful to observe the start of
   * an animation.
   */
  public void show(boolean wait)
  {
    gidlet.display.setCurrent(this);
    if (wait)
    {
      waitVisible = true;
      Timer t = new Timer();
      t.schedule(new MyTimerTask(), 5000);
      while (waitVisible && !isShown())
        Gidlet.delay(100);
      t.cancel();
    }
  }

  /**
   * Returns current content.
   */
  public String getText()
  {
    return getString();
  }

  /**
   * Writes given text replacing the previous content.
   * If the length of the text exceeds the maximal size granted
   * by the platform, the text is truncated.
   * If the text exceeds the height if the display, the
   * first line is on the top and the text may be scrolled using a small scroll
   * icon. Lines are separated by the plattform's line separator. If the length
   * of a line exceeds the width of the display, the line wraps at word boundaries.<br>
   * Newline character '\n' is supported to force a line wrap.
   */
  public void setText(String text)
  {
    setString(text.substring(0, text.length() > maxSize ? maxSize : text.length()));
  }

 /**
   * Appends given text to current content.
   * If the length of the total text exceeds the maximal size granted
   * by the platform, the text is truncated.
   * If the text exceeds the height if the display, the
   * first line is on the top and the text may be scrolled using a small scroll
   * icon. Lines are separated by the plattform's line separator. If the length
   * of a line exceeds the width of the display, the line wraps at word boundaries.<br>
   * Newline character '\n' is supported to force a line wrap.
   */
  public void appendText(String text)
  {
    String s = getString();
    s += text;
    setString(s.substring(0, text.length() > maxSize ? maxSize : text.length()));
  }

  /**
   * Adds an OK soft button with label "OK". When pressed, the
   * Gidlet's callback notification doOk() is called unless you
   * register your own CommandListener.
   */
  public void addOkButton()
  {
    addOkButton("OK");
  }

  /**
   * Adds an OK soft button with given label. When pressed, the
   * Gidlet's callback notification doOk() is called unless you
   * register your own CommandListener.
   */
  public void addOkButton(String label)
  {
    okCommand = new Command(label, Command.OK, 1);
    addCommand(okCommand);
  }

  /**
   * Removes Exit soft button.
   * Gidlet's callback notification doExit() is disabled.
   */
  public void removeExitButton()
  {
   removeCommand(exitCommand);
  }

  /**
   * Removes Ok soft button.
   * Gidlet's callback notification doOk() is disabled.
   */
  public void removeOkButton()
  {
   removeCommand(okCommand);
  }
}





