// Playground.java

/*
  This software is part of the Gidlet framework.
  It is Open Source Free Software, so you may
    - run the code for any purpose
    - study how the code works and adapt it to your needs
    - integrate all or parts of the code in your own programs
    - redistribute copies of the code
    - improve the code and release your improvements to the public
  However the use of the code is entirely your responsibility.
 */

package ch.aplu.gidlet;

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import java.io.*;
import java.util.*;
import ch.aplu.gidlet.*;

/**
 * Graphics window of full screen width and height (client drawing area),
 * exclusively used as canvas for the class Turtle.
 * <br><br>
 *
 * A current graph position is used, that remembers the end position of the last
 * drawing process and where new drawings start.<br><br>
 *
 * A soft button 'Exit' is shown. Hitting the corresponding key will
 * call the Gidlet's doExit(). If you do not need this button,
 * invoke removeExitButton().<br>
 * If you implement your own CommandListener and call setCommandListener(), the internal
 * CommandListener is disabled.<br>
 *
 * @see ch.aplu.gidlet.Turtle
 *
 */

public class Playground extends Canvas
{
// ------------------ Inner class MyCommandListener -------------
private class MyCommandListener implements CommandListener
{
  public void commandAction(Command c, Displayable d)
  {
    if (c == exitCommand)
      gidlet.doExit();
    if (okCommand != null && c == okCommand)
    {
      gidlet.doOk();
      gidlet.doOk(d);
      if (gidlet.isSleeping())
        gidlet.wakeUp();
    }
  }
}

// ------------------ Inner class TimerTask ---------------------
private class MyTimerTask extends TimerTask
{
  public void run()
  {
    waitVisible = false;
  }
}

// ------------------ Inner class TimerTask ---------------------
private class TextItem
{
  public String text;
  public int x;
  public int y;
  public int color;

  TextItem(String text, int x, int y, int color)
  {
    this.text = text;
    this.x = x;
    this.y = y;
    this.color = color;
  }
}

// ------------------ End of inner classes ----------------------

  private Gidlet gidlet;
  private Graphics turtleBuf;
  private Image turtleImage;
  private Graphics traceBuf;
  private Image traceImage;

  private String title;
  private int xPix;
  private int yPix;
  private double xmin;
  private double xmax;
  private double ymin;
  private double ymax;
  private double range;
  private int yFloor;
  private int xCurrent;
  private int yCurrent;
  private double a, b, c, d;
  private int width;
  private int height;
  private int bgColor;
  private int turtlePenColor;
  private int tracePenColor;
  private int titleColor;
  private boolean waitVisible;
  private Turtle turtle;
  private int turtleColor;
  private boolean visible;
  private boolean isShown;
  private Hashtable textLabels = new Hashtable();

  private KeyListener keyListener = null;
  private Command exitCommand = new Command("Exit", Command.EXIT, 1);
  private Command okCommand = null;

  protected Playground(Turtle turtle, String title,
                       double range,
                       int bgColor, int turtleColor, int penColor, int titleColor,
                       boolean visible, boolean isShown)
  {
    super();
    this.gidlet = Gidlet.myGidlet;
    this.turtle = turtle;
    this.title = title;
    this.range = range;
    this.xmin = -range;
    this.xmax = range;
    this.ymin = -range;
    this.ymax = range;
    this.turtleColor = turtleColor;
    this.tracePenColor = penColor;
    this.bgColor = bgColor;
    this.titleColor = titleColor;
    this.visible = visible;
    this.isShown = isShown;
    if (Thread.currentThread().getName().equals("Gidlet-Thread"))
      init();
    else // defer initialization
      (gidlet.displays).addElement(this);
  }

  /**
   * Initializes Playground. Called by main() if Playground is created before main() is called.
   */
  protected void init()
  {
    addCommand(exitCommand);
    setCommandListener(new MyCommandListener());

    height = getHeight();
    width = getWidth();

    turtleImage = Image.createImage(width, height);
    turtleBuf = turtleImage.getGraphics();
    traceImage = Image.createImage(width, height);
    traceBuf = traceImage.getGraphics();

    window(xmin, xmax, ymin, ymax);
    clear(true); // init xCurrent, yCurrent

    traceBuf.setColor(titleColor);
    traceBuf.drawString(title, getWidth()/2, 5,
                        Graphics.TOP | Graphics.HCENTER);
    turtleBuf.drawImage(traceImage, 0, 0, 0);

    traceBuf.setColor(tracePenColor);
    traceBuf.setClip(0, yFloor, width-1, height-1);
    turtleBuf.setClip(0, yFloor, width-1, height-1);

    Font font = Font.getFont(Font.FACE_SYSTEM, Font.STYLE_PLAIN, Font.SIZE_MEDIUM);
    traceBuf.setFont(font);

    turtle.shape = new TurtleShape(this, 20, 30);
    turtle.shape.setRange(range);
    turtle.shape.draw(0, 0, 0, turtleColor, visible, true);
    if (isShown)
      show(true);
  }

  /**
   * Shows the form display if it is not yet visible or
   * it was previously hidden by another display.
   * Because it is not garanteed that the diplay is actually visible when
   * Display.setCurrent() returns, wait for a maximum of 5 s for the display
   * to get shown.
   *
   */
  public void show()
  {
    show(true);
  }

 /**
   * Shows the graphics display if it was previously hidden by another display
   * or not yet displayed.<br>
   * If wait = true, block until the display becomes actually visible, up
   * to a maximum auf 5 seconds. This may be useful to observe the start of
   * an animation.
   */
  protected void show(boolean wait)
  {
    gidlet.display.setCurrent(this);
    if (wait)
    {
      waitVisible = true;
      Timer t = new Timer();
      t.schedule(new MyTimerTask(), 5000);
      while (waitVisible && !isShown())
        Gidlet.delay(100);
      t.cancel();
    }
  }

  /**
   * Sets color of drawing pen in trace buffer.
   * Returns previous value.
   */
  protected int setTracePenColor(int color)
  {
    int oldColor = tracePenColor;
    traceBuf.setColor(color);
    tracePenColor = color;
    return oldColor;
  }

  /**
   * Clears the graphics buffers (fully paint with background color)
   * and draws a surrounding white rectangle.
   * If init = true, set the current graph position to window coordiantes (0, 0).
   */
  protected void clear(boolean init)
  {
    traceBuf.setColor(bgColor);
    traceBuf.fillRect(0, 0, width, height);
    traceBuf.setColor(Gidlet.WHITE);
    traceBuf.drawRect(0, yFloor,
                      width-1, width-1);
    traceBuf.setColor(tracePenColor);

    turtleBuf.drawImage(traceImage, 0, 0, 0);
    repaint();
    if (init)
    {
      xCurrent = toPixelX(0);
      yCurrent = toPixelY(0);
    }
  }

  /**
   * Shows the given text with given color left adjusted at given x, y pixel coordinate.
   * Index is used to reference the text for erasing.
   */
  protected void addLabel(String text, int x, int y, int index, int color)
  {
    TextItem item = new TextItem(text, x, y, color);
    textLabels.put(new Integer(index), item);
    drawLabels();
  }

  protected void drawLabels()
  {
    for (Enumeration e = textLabels.keys(); e.hasMoreElements();)
    {
      Integer key = (Integer)e.nextElement();
      TextItem item = (TextItem)textLabels.get(key);
      turtleBuf.setColor(item.color);
      turtleBuf.drawString(item.text, item.x, item.y, Graphics.TOP | Graphics.LEFT);
      turtleBuf.setColor(turtlePenColor);
    }
  }

  /**
   * Removes text with given index.
   */
  protected void removeLabel(int index)
  {
    textLabels.remove(new Integer(index));
  }

  /**
   * Draws a line with given pixel coordinates into trace buffer
   * and sets the graph position to the endpoint.
   */
  protected void _line(int x1, int y1, int x2, int y2)
  {
    traceBuf.drawLine(x1, y1, x2, y2);
    xCurrent = x2;
    yCurrent = y2;
  }

  /**
   * Draws a line into trace buffer with given window coordinates into given buffer
   * and sets the graph position to the endpoint.
   */
  protected void line(double x1, double y1, double x2, double y2)
  {
    int xStart = toPixelX(x1);
    int yStart = toPixelY(y1);
    int xEnd = toPixelX(x2);
    int yEnd = toPixelY(y2);
    _line(xStart, yStart, xEnd, yEnd);
  }

  /**
   * Draws a line into trace buffer from current graph position to given pixel coordinates and
   * sets the graph position to the endpoint.
   */
  protected void _draw(int x, int y)
  {
    _line(xCurrent, yCurrent, x, y);
  }

  /**
   * Draws a line into trace buffer from current graph position to given window coordinates and
   * sets the graph position to the endpoint.
   */
  protected void draw(double x, double y)
  {
    _draw(toPixelX(x), toPixelY(y));
  }

  /**
   * Sets the current graph position into trace buffer to given pixel coordinates
   * (without drawing anything).
   */
  protected void _move(int x, int y)
  {
    xCurrent = x;
    yCurrent = y;
  }

  /**
   * Sets the current graph position in trace buffer to given window coordinates
   * (without drawing anything).
   */
  protected void move(double x, double y)
  {
    _move(toPixelX(x), toPixelY(y));
  }

  /**
   * Draws a triangle with given corners in pixel coordintates.
   * The graph position is unchanged.
   */
  protected void _triangle(int x1, int y1,
                        int x2, int y2,
                        int x3, int y3)
  {
    int oldX = xCurrent;
    int oldY = yCurrent;
    _move(x1, y1);
    _draw(x2, y2);
    _draw(x3, y3);
    _draw(x1, y1);
    xCurrent = oldX;
    yCurrent = oldY;
  }

  /**
   * Draws a triangle with given corners in window coordintates.
   * The graph position is unchanged.
   */
  protected void triangle(double x1, double y1,
                       double x2, double y2,
                       double x3, double y3)
  {
    _triangle(toPixelX(x1), toPixelY(y1),
              toPixelX(x2), toPixelY(y2),
              toPixelX(x3), toPixelY(y3));
  }

  /**
   * Draws a filled triangle with given corners in pixel coordintates into trace buffer
   * using the given color.
   * The graph position is unchanged.
   */
  protected void _fillTriangle(int x1, int y1,
                            int x2, int y2,
                            int x3, int y3,
                            int color)
  {
    traceBuf.setColor(color);
    traceBuf.fillTriangle(x1, y1, x2, y2, x3, y3);
    traceBuf.setColor(tracePenColor);
  }

  /**
   * Draws a filled triangle with given corners in window coordintates
   * using the given color.
   * The graph position is unchanged.
   */
  protected void fillTriangle(double x1, double y1,
                           double x2, double y2,
                           double x3, double y3,
                           int color)
  {
    _fillTriangle(toPixelX(x1), toPixelY(y1),
                  toPixelX(x2), toPixelY(y2),
                  toPixelX(x3), toPixelY(y3), color);
  }

  /**
   * Draws a rectangle with given opposite corners
   * in pixel coordinates into trace buffer.
   * ulx: upper left x
   * uly: upper left y
   * lrx: lower right x
   * lry: lower right y
   * The graph position is unchanged.
   */
  protected void _rectangle(int ulx, int uly, int lrx, int lry)
  {
    traceBuf.drawRect(ulx, uly, lrx - ulx, lry - uly);
  }

  /**
  * Draws a rectangle with given opposite corners
  * in window coordinates.
  * The graph position is unchanged.
  */
  protected void rectangle(double x1, double y1, double x2, double y2)
  {
    _rectangle(toPixelX(x1), toPixelY(y2), toPixelX(x2), toPixelY(y1));
  }

  /**
   * Draws a filled rectangle with given opposite corners
   * in pixel coordinates into trace buffer using the given color.
   * ulx: upper left x
   * uly: upper left y
   * lrx: lower right x
   * lry: lower right y
   * The graph position is unchanged.
   */
  protected void _fillRectangle(int ulx, int uly, int lrx, int lry, int color)
  {
    traceBuf.setColor(color);
    traceBuf.fillRect(ulx, uly, lrx - ulx, lry - uly);
    traceBuf.setColor(tracePenColor);
  }

  /**
   * Draws a filled rectangle with given opposite corners
   * in window coordinates using the given color.
   * The graph position is unchanged.
   */
  protected void fillRectangle(double x1, double y1, double x2, double y2, int color)
  {
    _fillRectangle(toPixelX(x1), toPixelY(y2), toPixelX(x2), toPixelY(y1), color);
  }

  /**
   * Sets window coordinates.
   * left_x, right_x, bottom_y, top_y
   */
  protected void window(double xmin, double xmax, double ymin, double ymax)
  {
    double ytop = ymin + getAspectRatio()*(ymax - ymin);
    this.xmin = xmin;
    this.xmax = xmax;
    this.ymin = ymin;
    this.ymax = ytop;
    xPix = width - 1;
    yPix = height - 1;

    a = xPix / (this.xmax - this.xmin);
    b = xPix * this.xmin / (this.xmin - this.xmax);
    c = yPix / (this.ymin - this.ymax);
    d = yPix * this.ymax / (this.ymax - this.ymin);

    xCurrent = toPixelX(0);
    yCurrent = toPixelY(0);
    yFloor = toPixelY(ymax);
   }

  /**
   * Converts window coordinates to pixel coordinates (horizontal).
   */
  protected int toPixelX(double windowX)
  {
    return rint(a * windowX + b);     // Rounding
  }

  /**
   * Converts window coordinates to pixel coordinates (vertical).
   */
  protected int toPixelY(double windowY)
  {
    return rint(c * windowY + d);     // Rounding
  }

  /**
   * Converts window coordinates increment to pixel coordinates increment (horizontal).
   * Increment is always positive.
   */
  protected int toPixelWidth(double windowWidth)
  {
    return rint(Math.abs(a * windowWidth));
  }

  /**
   * Converts window coordinates increment to pixel coordinates increment (vertical).
   * Increment is always positive.
   */
  protected int toPixelHeight(double windowHeight)
  {
    return rint(Math.abs(c * windowHeight));
  }

  /**
   * Converts pixel coordinates to window coordinates (horizontal).
   */
  protected double toWindowX(int userX)
  {
    return (userX - b) / a;
  }

  /**
   * Converts pixel coordinates to window coordinates (vertical).
   */
  protected double toWindowY(int userY)
  {
    return (userY - d) / c;
  }

  /**
   * Converts pixel coordinates increment to window coordinates increment (horizontal).
   * Increment is always positive.
   */
  protected double toWindowWidth(int pixelWidth)
  {
    return Math.abs(pixelWidth / a);
  }

  /**
   * Converts pixel coordinates increment to window coordinates increment (vertical).
   * Increment is always positive.
   */
  protected double toWindowHeight(int pixelHeight)
  {
    return Math.abs(pixelHeight / c);
  }

 /**
  * Returns the ratio of canvas height to width.
  */
  protected double getAspectRatio()
  {
    return (height / (double)width);
  }

  /**
   * Returns pixel coordinate x of current graph position.
   */
  protected int _getPosX()
  {
    return xCurrent;
  }

  /**
   * Returns window coordinate x of current graph position.
   */
  protected double getPosX()
  {
    return toWindowX(xCurrent);
  }

  /**
   * Returns pixel coordinate y of current graph position.
   */
  protected int _getPosY()
  {
    return yCurrent;
  }

  /**
   * Returns window coordinate y of current graph position.
   */
  protected double getPosY()
  {
    return toWindowY(yCurrent);
  }

 /**
  * Returns the graphics turtle buffer.
  */
  protected Graphics getTurtleBuffer()
  {
    return turtleBuf;
  }

 /**
  * Returns the graphics turtle buffer.
  */
  protected Graphics getTraceBuffer()
  {
    return traceBuf;
  }

 /**
  * Returns the turtle image.
  */
  protected Image getTurtleImage()
  {
    return turtleImage;
  }

 /**
  * Returns the trace image.
  */
  protected Image getTraceImage()
  {
    return traceImage;
  }

  /**
   * Registers an KeyListener to get a notification when a button is clicked.
   */
  protected void addKeyListener(KeyListener keyListener)
  {
    this.keyListener = keyListener;
  }

  /**
   * Removes a registered KeyListener.
   */
  protected void removeKeyListener()
  {
    keyListener = null;
  }

  /**
   * Puts the PNG image into turtle buffer  from given file url at given pixel coordinates
   * (specifies upper left corner of image).
   * Returns true if successful.
   */
  protected boolean _image(String imageUrl, int x, int y)
  {
    InputStream is = getClass().getResourceAsStream(imageUrl);
    Image img;
    try
    {
      img = Image.createImage(is);  // Blocking until read and decoded
      if (is != null)
        is.close();
    }
    catch (IOException ex)
    {
      return false;
    }
    int imgHeight = img.getHeight();
    int imgWidth = img.getWidth();
    turtleBuf.drawImage(img, x, y, Graphics.TOP | Graphics.LEFT);
    return true;
  }

  /**
   * Puts the PNG image into turtle buffer from given file url at given window coordinates
   * (specifies upper left corner of image).
   * Returns true if successful.
   */
  protected boolean image(String imageUrl, double x, double y)
  {
    return _image(imageUrl, toPixelX(x), toPixelY(y));
  }

  /**
   * Returns the width (horizontal size) of the PNG image from the given url (in pixel coordinates).
   * Returns 0, if PNG image is invalid.
   */
  protected int _imageWidth(String imageUrl)
  {
    InputStream is = getClass().getResourceAsStream(imageUrl);
    Image img;
    try
    {
      img = Image.createImage(is);
      if (is != null)
        is.close();
    }
    catch (IOException ex)
    {
      return 0;
    }
    return img.getWidth();
  }

  /**
   * Returns the width (horizontal size) of the PNG image from the given url (in window coordinates).
   * Returns 0, if PNG image is invalid.
   */
  protected double imageWidth(String imageUrl)
  {
    int rc = _imageWidth(imageUrl);
    if (rc == 0)
      return 0;
    else
      return toWindowWidth(rc);
  }

  /**
   * Returns the height (vertical size) of the PNG image from the given url (in pixel coordinates).
   * Returns 0, if PNG image is invalid.
   */
  protected int _imageHeight(String imageUrl)
  {
    InputStream is = getClass().getResourceAsStream(imageUrl);
    Image img;
    try
    {
      img = Image.createImage(is);
      if (is != null)
       is.close();
    }
    catch (IOException ex)
    {
      return 0;
    }
    return img.getHeight();
  }

  /**
   * Returns the height (vertical size) of the PNG image from the given url (in window coordinates).
   * Returns 0, if PNG image is invalid.
   */
  protected double imageHeight(String imageUrl)
  {
    int rc = _imageHeight(imageUrl);
    if (rc == 0)
      return 0;
    else
      return toWindowHeight(rc);
  }

  /**
   * Returns width of screen in pixel coordinates.
   * Pixel coordinates 0 up to this value (inclusive)
   * Number of pixels: this value + 1
   */
  protected int _screenWidth()
  {
    return xPix;
  }

  /**
   * Returns number of pixels in horizontal direction.
   */
  protected int screenResolutionX()
  {
    return width;
  }

  /**
   * Returns width of screen in window coordinates.
   */
  protected double screenWidth()
  {
    return toWindowWidth(xPix);
  }

  /**
   * Returns height of screen in pixel coordinates.
   * Pixel coordinates 0 up to this value (inclusive)
   * Number of pixels: this value + 1
   */
  protected int _screenHeight()
  {
    return yPix;
  }

  /**
   * Returns number of pixels in vertical direction.
   */
  protected int screenResolutionY()
  {
    return height;
  }

  /**
   * Returns height of screen in window coordinates.
   */
  protected double screenHeight()
  {
    return toWindowHeight(yPix);
  }

  /**
   * Rounding doubles
   */
  private int rint(double x)
  {
    if (x > 0)
      return (int)(x + 0.5);
    else
      return (int)(x - 0.5);
   }

  /**
   * Used for rendering canvas.
   */
  protected void paint(Graphics g)
  {
    g.drawImage(turtleImage, 0, 0, 0);
  }

  /**
   * Used to get key events.
   */
  protected void keyPressed(int keyCode)
  {
    if (keyListener != null)
      keyListener.keyPressed(keyCode);
  }

  /**
   * Used to get key events.
   */
  protected void keyReleased(int keyCode)
  {
    if (keyListener != null)
      keyListener.keyReleased(keyCode);
  }

  /**
   * Used to get key events.
   */
  protected void keyRepeated(int keyCode)
  {
    if (keyListener != null)
      keyListener.keyRepeated(keyCode);
  }

  /**
   * Adds an OK soft button with label "OK". When pressed, the
   * Gidlet's callback notification doOk() is called unless you
   * register your own CommandListener.
   */
  public void addOkButton()
  {
    addOkButton("OK");
  }

  /**
   * Adds an OK soft button with given label. When pressed, the
   * Gidlet's callback notification doOk() is called unless you
   * register your own CommandListener.
   */
  public void addOkButton(String label)
  {
    okCommand = new Command(label, Command.OK, 1);
    addCommand(okCommand);
  }

  /**
   * Removes Exit soft button.
   * Gidlet's callback notifications doExit() is disabled.
   */
  public void removeExitButton()
  {
   removeCommand(exitCommand);
  }

  /**
   * Removes OK soft button.
   * Gidlet's callback notification doOk() is disabled.
   */
  public void removeOkButton()
  {
   removeCommand(okCommand);
  }

}


