// TurtleShape.java

/*
  This software is part of the Gidlet framework.
  It is Open Source Free Software, so you may
    - run the code for any purpose
    - study how the code works and adapt it to your needs
    - integrate all or parts of the code in your own programs
    - redistribute copies of the code
    - improve the code and release your improvements to the public
  However the use of the code is entirely your responsibility.
*/

/*
Wrapping the turtle shape:

Only window 5 is visible, but may be drawn in other windows, due to automatic
clipping. For correct wrapping turtle shape should be drawn in all 9 windows.

Some optimations:

Draw only n other windows than 5 if close to the border:
We have 4 borders: left, right, bottom, top.
If close to left: draw in 1, 4, 5
If close to right: draw in 3, 6, 9
If close to bottom: draw in 1, 2, 3
If close to top: draw in 7, 8, 9
If more than condition is true, draw take the union.


 ---- ---- ----
| 1  | 2  | 3  |
|    |    |    |
 ---- ---- ----
| 4  | 5  | 6  |
|    | xx |    | xx: visible window
 ---- ---- ----
| 7  | 8  | 9  |
|    |    |    |
 ---- ---- ----
*/

package ch.aplu.gidlet;

import ch.aplu.gidlet.*;
import javax.microedition.lcdui.*;
import java.util.*;

class TurtleStamp
{
   // Position (anchor point) of turte that is completely or partially shown
   // Anchor point inside borders or in small band outside borders
  public double x, y;

  public double phi;
  public int turtleColor;

  public TurtleStamp(double x, double y, double phi, int color)
  {
    this.x = x;
    this.y = y;
    this.phi = phi;
    this.turtleColor = color;
  }
}


class TurtleShape
{

  private Playground pg;
  private double a, b, h;
  private double phi;
  private double range;
  private boolean isRepaint = true;
  protected Vector stampedTurtles = new Vector();  // Turtle must access

  protected TurtleShape(Playground pg, double a, double b)
  {
    this.pg = pg;
    this.a = a;
    this.b = b;
    this.h = Math.sqrt(b*b - a*a/4);
  }

  protected void setRange(double range)
  {
    this.range = range;
  }

  protected void setRepaint(boolean b)
  {
    this.isRepaint = b;
  }

  protected boolean getRepaint()
  {
    return isRepaint;
  }

  protected void draw(double x, double y, double phi, int turtleColor,
                      boolean isVisible, boolean isWrap)
  {
    Graphics turtleBuf = pg.getTurtleBuffer();
    Image traceImage = pg.getTraceImage();

    // Copy trace buffer into turtle buffer
    turtleBuf.drawImage(traceImage, 0, 0, 0);

    // Copy stamped turtles into turtle buffer
    for (int i = 0; i < stampedTurtles.size(); i++)
    {
      TurtleStamp ts = (TurtleStamp)stampedTurtles.elementAt(i);
      if (isWrap)
        drawWrappedShapes(ts.x, ts.y, ts.phi, ts.turtleColor);
      else
        drawShape(ts.x, ts.y, ts.phi, ts.turtleColor);
    }

    if (isVisible)
    {
      if (isWrap)
      {
        drawWrappedShapes(x, y, phi, turtleColor);
      }
      else // no wrap
        drawShape(x, y , phi, turtleColor);
    }
    pg.drawLabels();
    if (isRepaint)
      pg.repaint();
  }

  private void drawShape(double x, double y, double phi, int color)
  {
    Graphics turtleBuf = pg.getTurtleBuffer();

    double p = phi / 180 * Math.PI;

    double p1x = x - a/2 * Math.cos(p);
    double p1y = y + a/2 * Math.sin(p);

    double p2x = x + a/2 * Math.cos(p);
    double p2y = y - a/2 * Math.sin(p);

    double p3x = x + h * Math.sin(p);
    double p3y = y + h * Math.cos(p);

    int _p1x = pg.toPixelX(p1x);
    int _p1y = pg.toPixelY(p1y);
    int _p2x = pg.toPixelX(p2x);
    int _p2y = pg.toPixelY(p2y);
    int _p3x = pg.toPixelX(p3x);
    int _p3y = pg.toPixelY(p3y);

    turtleBuf.setColor(color);
    turtleBuf.fillTriangle(_p1x, _p1y, _p2x, _p2y, _p3x, _p3y);
  }

  protected double wrapToRange(double x)
  {
    int xInt = (int)(100.0 * x);   // Multiply by 100 to get better resolution
    int rInt = (int)(100.0 * range);
    int remainder;

    if (xInt >= 0)
      remainder = (xInt + rInt) % (2*rInt) - rInt;
    else
    {
      int xBorder = (-xInt + rInt) % (2*rInt);
      if (xBorder == 0)  // just  on border
        remainder = -rInt;  // remain on this border
      else
        remainder = -xBorder + rInt;
    }
    return (double)(remainder) / 100.0;

  }

  private void drawWrappedShapes(double x, double y, double phi, int turtleColor)
  {
    double _x = wrapToRange(x);
    double _y = wrapToRange(y);
    drawShape(_x, _y , phi, turtleColor);

    boolean do1 = false;
    boolean do2 = false;
    boolean do3 = false;
    boolean do4 = false;
    boolean do6 = false;
    boolean do7 = false;
    boolean do8 = false;
    boolean do9 = false;

    if (range - _x < h)  // near right border
      do1 = do4 = do7 = true;

    if (range + _x < h)  // near left border
      do3 = do6 = do9 = true;

    if (range - _y < h) // near top border
      do7 = do8 = do9 = true;

    if (range + _y < h) // near bottom border
      do1 = do2 = do3 = true;

    if (do1)
      drawShape(_x - 2*range, _y + 2*range , phi, turtleColor);
    if (do2)
      drawShape(_x,           _y + 2*range , phi, turtleColor);
    if (do3)
      drawShape(_x + 2*range, _y + 2*range , phi, turtleColor);
    if (do4)
      drawShape(_x - 2*range, _y ,           phi, turtleColor);
    if (do6)
      drawShape(_x + 2*range, _y,            phi, turtleColor);
    if (do7)
      drawShape(_x - 2*range, _y - 2*range , phi, turtleColor);
    if (do8)
      drawShape(_x,           _y - 2*range , phi, turtleColor);
    if (do9)
      drawShape(_x + 2*range, _y - 2*range , phi, turtleColor);
  }
}



