// SoundSensor.java

/*
This software is part of the NxtJLib library.
It is Open Source Free Software, so you may
- run the code for any purpose
- study how the code works and adapt it to your needs
- integrate all or parts of the code in your own programs
- redistribute copies of the code
- improve the code and release your improvements to the public
However the use of the code is entirely your responsibility.
*/

package ch.aplu.nxt;

import ch.aplu.nxt.platform.*;

/**
 * Class that represents a sound sensor.
 */
public class SoundSensor extends Sensor implements SharedConstants
{
  private interface SensorState // Simulate enum for J2ME compatibility
  {
    int QUIET = 0;
    int LOUD = 1;
  }

// -------------- Inner class SoundSensorThread ------------
  private class SoundSensorThread extends NxtThread
  {
    private volatile boolean isRunning = false;

    private SoundSensorThread()
    {
      if (NxtRobot.getDebugLevel() >= DEBUG_LEVEL_LOW)
        DebugConsole.show("DEBUG: SoundSensorThread created (port: " +
          getPortLabel() + ")");
    }

    public void run()
    {
      if (NxtRobot.getDebugLevel() >= DEBUG_LEVEL_LOW)
        DebugConsole.show("DEBUG: SoundSensorThread " +
          Thread.currentThread().getName() +
          " started (port: " + getPortLabel() + ")");
      isRunning = true;
      while (isRunning)
      {
        if (soundListener != null)
        {
          Tools.delay(pollDelay);
          int level = getLevel();
          if (state == SensorState.QUIET && level > triggerLevel)
          {
            if (inCallback)
            {
              if (NxtRobot.getDebugLevel() >= DEBUG_LEVEL_LOW)
                DebugConsole.show("DEBUG: Sound event 'loud' (port: " +
                  getPortLabel() + ") -------- rejected: Other callback underway!");
            }
            else
            {  
              inCallback = true;
              if (NxtRobot.getDebugLevel() >= DEBUG_LEVEL_LOW)
                DebugConsole.show("DEBUG: Sound event 'loud' (port: " +
                  getPortLabel() + ")");
              soundListener.loud(getPort(), level);
              state = SensorState.LOUD;
              inCallback = false;
            }
          }
          if (state == SensorState.LOUD && level <= triggerLevel)
          {
            if (inCallback)
            {
              if (NxtRobot.getDebugLevel() >= DEBUG_LEVEL_LOW)
                DebugConsole.show("DEBUG: Sound event 'quiet' (port: " +
                  getPortLabel() + ") -------- rejected: Other callback underway!");
            }
            else
            {  
              inCallback = true;
              if (NxtRobot.getDebugLevel() >= DEBUG_LEVEL_LOW)
                DebugConsole.show("DEBUG: Sound event 'quiet' (port: " +
                  getPortLabel() + ")");
              soundListener.quiet(getPort(), level);
              state = SensorState.QUIET;
              inCallback = false;
            }
          }
        }
      }
    }

    private void stopThread()
    {
      isRunning = false;
      try
      {
        joinX(500);
      }
      catch (InterruptedException ex)
      {
      }
      if (NxtRobot.getDebugLevel() >= DEBUG_LEVEL_LOW)
        if (isAlive())
          DebugConsole.show("DEBUG: SoundSendorThread stopping failed (port: " +
            getPortLabel() + ")");
        else
          DebugConsole.show("DEBUG: SoundSensorThread successfully stopped (port: " +
            getPortLabel() + ")");
    }
  }
// -------------- End of inner classes -----------------------
  private SoundListener soundListener = null;
  private SoundSensorThread sst;
  private int state = SensorState.QUIET;
  private int triggerLevel;
  private final int pollDelay;
  private volatile static boolean inCallback = false;

  /**
   * Creates a sensor instance connected to the given port.
   * @param port the port where the sensor is plugged-in
   */
  public SoundSensor(SensorPort port)
  {
    super(port);
    sst = new SoundSensorThread();
    NxtProperties props = NxtRobot.getProperties();
    pollDelay = props.getIntValue("SoundSensorPollDelay");
  }

  /**
   * Creates a sensor instance connected to port S1.
   */
  public SoundSensor()
  {
    this(SensorPort.S1);
  }
  
  // Called by connect() if part is alread added.
  // Called by addPart() if already connected
  protected void init()
  {
    if (NxtRobot.getDebugLevel() >= DEBUG_LEVEL_MEDIUM)
      DebugConsole.show("DEBUG: SoundSensor.init() called (Port: " +
        getPortLabel() + ")");

    setTypeAndMode(SOUND_DB, PCTFULLSCALEMODE);
  }

  protected void cleanup()
  {
    if (NxtRobot.getDebugLevel() >= DEBUG_LEVEL_MEDIUM)
      DebugConsole.show("DEBUG: SoundSensor.cleanup() called (Port: " +
        getPortLabel() + ")");
    if (sst != null)
      sst.stopThread();
  }

  /**
   * Registers the given sound listener for the given trigger level.
   * If the sound thread is not yet started, start it now.
   * @param soundListener the SoundListener to become registered.
   * @param triggerLevel the trigger level where the callback is triggered
   */
  public void addSoundListener(SoundListener soundListener, int triggerLevel)
  {
    this.soundListener = soundListener;
    this.triggerLevel = triggerLevel;
    if (!sst.isAlive())
      startSoundThread();
  }

  protected void startSoundThread()
  {
    sst.start();
  }

  protected void stopSoundThread()
  {
    if (sst.isAlive())
      sst.stopThread();
  }

  private int getLevel()
  {
    if (robot == null || !robot.isConnected())
      return -1;
    return readScaledValue();
  }

  /**
   * Polls the sensor.
   * @return the current value the sensor reported: 0 (quiet) .. 150 (loud)
   */
  public int getValue()
  {
    checkConnect();
    return readScaledValue();
  }

  private void checkConnect()
  {
    if (robot == null)
      new ShowError("SoundSensor (port: " + getPortLabel() +
        ") is not a part of the NxtRobot.\n" +
        "Call addPart() to assemble it.");
    if (!robot.connect())
      PlatformTools.putSleep();
  }
}
